// ILI9341.js - emulating module for ILI9341 (and other) controllers

// 20190927 (c) allObjects

// Uses 'Modules._eCache("<moduleName>",exports) to emulate caching on 'upload'

(function() { var exports = {}; // create (emulating) module load context

var G = function () {
 this._eCanvas = null;
 this._eClr = [255, 255, 255];
 this._eFontSize = 20;
 this._eFontFamily = "Arial, Helv";
};

G.prototype.clear = function() {
    var c2 = this._eCanvas.getContext("2d");
    var w = this._eCanvas.width, h = this._eCanvas.height;
    c2.fillStyle = "#000000";
    c2.strokeStyle = "#000000";
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(0,0);
    c2.lineTo(w-1,0); c2.lineTo(w-1,h-1); c2.lineTo(0,h-1);
    c2.closePath();
    c2.fill();
    return this;
  };
G.prototype.setColor = function(r,g,b) {
    this._eClr = "#" + this._eHex2(Math.round(r * 255)) 
        + this._eHex2(Math.round(g * 255)) + this._eHex2(Math.round(b * 255));
    return this;
  };
G.prototype.drawRect = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 2;
    c2.beginPath();
    c2.moveTo(x1,y1);
    c2.lineTo(x2,y1);
    c2.lineTo(x2,y2);
    c2.lineTo(x1,y2);
    c2.closePath();
    c2.stroke();
//    c2.strokeRect(x1,y1,x2-x1,y2-y1);
    return this;
  };
G.prototype.fillRect = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.fillRect(x1,y1,x2-x1+1,y2-y1+1);
    return this;
  };
G.prototype.drawPoly = function(vs,closed) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 2;
    c2.shadow
    c2.beginPath();
    c2.moveTo(vs[0], vs[1]);
    for (var i = 3; i < vs.length; i = i + 2) { c2.lineTo(vs[i - 1], vs[i]); }
    if (closed) { c2.lineTo(vs[0], vs[1]); }
    c2.stroke();
    return this;
  };
G.prototype.fillPoly = function(vs) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(vs[0], vs[1]);
    for (var i = 3; i < vs.length; i = i + 2) { c2.lineTo(vs[i-1], vs[i]); }
    c2.closePath();
    c2.stroke();
    c2.fill();
    return this;
  };
G.prototype.drawLine = function(x1,y1,x2,y2) {
    var c2 = this._eCanvas.getContext("2d");
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.lineWidth = 1;
    c2.beginPath();
    c2.moveTo(x1,y1);
    c2.lineTo(x2,y2);
    c2.stroke();
    return this;
  };
G.prototype.setFontVector = function(size) {
    this._eFontSize = size;
    return this;
  };
G.prototype.drawString = function(t,x,y) {
    var c2 = this._eCanvas.getContext("2d");
    c2.font = this._eFont();
    c2.fillStyle = this._eClr;
    c2.strokeStyle = this._eClr;
    c2.fillText(t,x,y + Math.round(this._eFontSize * 1.0));
    return this;
  };
G.prototype._eRegisterCanvas = function(_eCanvasId, w, h) {
    _eConsole.log("ILI9341._eRegisterCanvas() id:"+_eCanvasId,"w:"+w,"h"+h);
    this._eCanvas = document.getElementById(_eCanvasId);
    this._eCanvas.width = w;
    this._eCanvas.height = h;
    return this;
  };
G.prototype._eHex2 = function(n) {
    return ("0123456789abcdef".charAt(n/16)
          + "0123456789abcdef".charAt(n % 16) );
  };
G.prototype._eFont = function() { return ("" 
      + Math.round(this._eFontSize * 1.2) + "px " + this._eFontFamily);
  };

exports.connect = function(spi, b6, b8, b7, cb) {
    _eConsole.log("ILI9341.connect(); // incl. opt. _eRegisterCanvas();");
    var g =  new G(), args;
    if (args = Modules._eeArgs["ILI9341"]) {
      g._eRegisterCanvas(args[1],args[2],args[3]); }
    setTimeout(function(){ cb(); }, 100);
    return g;
  };

Modules._eCache("ILI9341", exports); })(); // load emulating module
