// TouchRD.js - emulating module for resistive touch screen direct controlled

// 20190912 - (c) allObjects

// Uses 'Modules._eCache("<moduleName>",exports) to emulate caching on 'upload'

// Interfacing a resistive touchscreen's X+, Y+, X-, and Y- directly with Espruino
// (for touchscreen breakout boards that do not have a touchscreen controller, 
// such as ADS7843, Espruino IS the touch screen controller using 4 ADC pins) 

(function() { var exports = {}; // create (emulating) module load context

var T = function(cb) {
  this.l = undefined; // listening ( undefined: never yet, false: not, true: yes  
  this.x = 0;
  this.y = 0;
  this.p = false; // previous touch
  this.t = false; // touching
  this.d = false; // touch down
  this.cb = cb;
  this._eCanvas = null;
  this._eXoff = 0;
  this._eYoff = 0;
};

T.prototype.listen = function(b) {
    _eConsole.log("TouchRD.listen();");
    if ((b || (b === undefined)) && (this.l === undefined)) { this._eSetupListeners(); }
    this.l = (b || (b === undefined));
  }
T.prototype._eRegisterCanvas = function(_eCanvasId, _eXoff, _eYoff) {
    _eConsole.log("TouchRD._eRegisterCanvas() id:"+_eCanvasId,"_eXoff:"+_eXoff,"_eYoff:"+_eYoff);
    this._eCanvas = document.getElementById(_eCanvasId);
    this._eXoff = _eXoff || 0;
    this._eYoff = _eYoff || 0;
    return this;
  };
T.prototype._eSetupListeners = function() {
    _eConsole.log("TouchRD._eSetupListeners();");
    var _this = this, c = this._eCanvas, e,
    h = function(evt,t) {
      if (!_this.l) return;
      e = (evt) ? evt : event;
      _this.x = e.pageX - c.offsetLeft - _this._eXoff;
      _this.y = e.pageY - c.offsetTop  - _this._eYoff;
      if (t !== undefined) {
        _this.d = (!_this.t && t);
        _this.t = t;
      } else {
        _this.d = false;
      }
      if (_this.t) {
        c.style.cursor = "pointer";
        _this.p = true;
        _this.cb(_this.x, _this.y);
      } else {
        if (_this.p) {
          _this.p = false;
          c.style.cursor = "auto";
          _this.cb();
        }
      }
    },
    doLogEvt = false, // false, // true,
    logEvt = function(en) { if (doLogEvt) { console.log(en); } },
    mouseDown = function(e) { logEvt("mousedown" ); h(e,true ); },
    mouseXY   = function(e) { logEvt("mousemove" ); h(e      ); },
    touchDown = function(e) { logEvt("touchstart"); h(e,true ); },
    touchXY   = function(e) { logEvt("touchmove" ); h(e      ); },
    touchUp   = function(e) { logEvt("touchend|touchcancel"); h(e,false); },
    mouseUp   = function(e) { logEvt("mouseup"   ); h(e,false); };
    c.addEventListener("mousedown" , mouseDown, false);
    c.addEventListener("mousemove" , mouseXY  , false);
    c.addEventListener("touchstart", touchDown, false);
    c.addEventListener("touchmove" , touchXY  , true );
    c.addEventListener("touchend"  , touchUp  , false);
    document.body.addEventListener("mouseup"    , mouseUp, false);
    document.body.addEventListener("touchcancel", touchUp, false);
    this.l = false;
  };

exports.connect = function(xP,xN,yP,yN,cb) {
    _eConsole.log("TouchRD.connect();");
    return new T(cb);
  };

Modules._eCache("TouchRD", exports); })(); // load emulating module
