// _eEspruinoJsShim.js - adding Espruino js built-in things to browser js context

// 20190927 - (c) allObject

// Used in Espruino cross development environment in Browser - extend as needed.
//
// Includes so far:
// -- getTime(), system time in SECONDS
// -- _eConsole singleton w/ .log(), .warn(), .error()
// -- require() 
// -- Modules - modules cache supporting require()


// === Support for emulation console (eConsole) logging (by default to standard console)
var _eConsole =
{ console: console // for opt. override w/ custom console (and keep _eLogNode null).
, consoleOn: true  // log to browser console as well
, _eLogNode: null  // if a (DOM) node (!= null), prepend as line to DOM node's body
, log:   function() { if (this._eLogNode) { this._eLog("I",arguments);
           } else { this.console.log.apply(  this.console, arguments); } }
, warn:  function() { if (this._eLogNode) { this._eLog("W",arguments);
           } else { this.console.warn.apply( this.console, arguments); } }
, error: function() { if (this._eLogNode) { this._eLog("E",arguments);
           } else { this.console.error.apply(this.console, arguments); } }
, _eLog: function(typ,args) { // log message type "I|W|E|orCustom", args / array
    var ss = [], sc = [], _eTs = ("000000" + (new Date().getTime() % 100000)).substr(-6)
      , cs = "|......................." // 24
      , cf = "|......................................." // 40
      , s, sfs, sfc;
    for (var i=0; i<args.length; i++) {
      if ((typeof (s = args[i]) == "string") && !(s.indexOf(cs))) {
        // console.log(s);
        sfs
        = (s.substr(-5) == "|btn|") ? cf.substr(0,20)
        : (s.substr(-5) == "|chk|") ? cf.substr(0,20)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,22)
        : (s.substr(-5) == "|sli|") ? cf.substr(0,20)
        : cf.substr(0,24);
        ss.push(sfs);
        sfc
        = (s.substr(-5) == "|btn|") ? cf.substr(0,24)
        : (s.substr(-5) == "|chk|") ? cf.substr(0,24)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,26)
        : (s.substr(-5) == "|rad|") ? cf.substr(0,24)
        : cf.substr(0,24);
        sc.push(sfc);
      } else {
        ss.push(s);
        sc.push(s);
      }
    }
    if (this.consoleOn) { sc.splice(0,0,[_eTs]); console.log.apply(console,sc); }
    this._eLogNode.innerHTML = _eTs + ": " + typ + ": " + ss.join(", ")
                             + "<br>" + this._eLogNode.innerHTML; }
, _eRegisterLogNode: function(id) { this._eLogNode = document.getElementById(id);
    this.log( "_eEmulationPath:"   + _eEmulationPath
            , "_eModulesPath:"     + _eModulesPath
            , "_eProjectsPath:"    + _eProjectsPath
            , "_eMinifiedDefault:" + _eMinifiedDefault
            ); }
};

// some dynamic hatml support for X dev html files
var _efs = []; // html fragments // re-init before use
function _ep(s) { _efs.push(s); } // add s to html fragments
function _ew() { document.writeln(_efs.join("")); _efs = []; } // write frags to html

// ===== Ports / Pins and Appliances

var Pin = function(id) { this.id = id; this.v = 0; this.ad = ""; }
  , SPI = function(id) { this.id = id; }
  ;
Pin.prototype.toString = function() { return (this.id + ":" + this.v); }
Pin.prototype["set"]   = function() {
    this.v = 1; log(this.id + ".set()" + ((this.ad) ? " - " + ad : "")); };
Pin.prototype["reset"] = function() {
    this.v = 0; log(this.id + ".reset()" + ((this.ad) ? " - " + ad : "")); };
SPI.prototype.toString = function() { return this.id; }
SPI.prototype.setup = function() { log(this.id + ".setup()"); };


// ===== functions

function getTime() {  // returns system time in seconds as floating number
  return (new Date().getTime()) / 1000;
}

function require(moduleName) {
  _eConsole.log('require("' + moduleName + '");');
  return Modules._eCached(moduleName);
}

// ==== Modules cache
var Modules =
{ modules: {}
, removeCached: function(id) {
    if (this.modules[id]) {
      delete this.modules[id];
    } else {
      _eConsole.warn("Modules.removeCached(): "
          + "module " + id + " not found or already removed from cache");
    }
  }
};

// ==== Support for emulated Modules cache
Modules._eCache = function(moduleName, module) {
    _eConsole.log('Modules._eCache("' + moduleName + '", ...);');
    this.modules[moduleName] = module;
  };
Modules._eCached = function(moduleName) {
    var module = this.modules[moduleName];
    if (!module) {
      if (!this.modules.hasOwnProperty(moduleName)) {
        this.modules[moduleName] = undefined;
      }
      _eConsole.error("ERROR: module with name " + moduleName + " not found.");
    }
    return module;
  };
// stores args for emulated modules at _emLoad()
Modules._eeArgs = {};
Modules._eeRegisterArgs = function(args) {
    var ss = []; for (var i=0; i<args.length; i++) { ss.push(args[i]); }
    _eConsole.log("Modules._eeRegisterArgs() - "
        + args.length + ": [ " +ss.join(", ") + " ]");
    this._eeArgs[args[0]] = args;
  };

// === Support for module load emulation (from modules folder);
// is part of emulated upload to board with the following difference:
// upload to board does parse recursively / 'nested' all code with
// regex before uploading to have all dependencies loaded. In emulation,
// This happens on html page load and is hardcoded / 'manually' *BEFORE*
// Espruino code by dynamically pulling module as src in html script tag. 
var exports;
function _emLoad(moduleName) {
  _eConsole.log('_emLoad("' + moduleName + '"...);');
  exports = {};
  document.writeln(['<scri','pt src="',
      _eModulesPath,moduleName,
      ((_eMinifiedDefault) ? ".min" : ""),'.js">',
      '</scri','pt>'].join(""));
  document.writeln(['<scri','pt>',
      'Modules._eCache(exports.mn, exports);',
      '</scri','pt>'].join(""));
}


// === Support for project load emulation (from projects folder);
// This allows to pull in / embed the Espruino upload code as is
function _epLoad(projectName) {
  _eConsole.log('_epLoad("' + projectName + '"...);');
  document.writeln(['<scri','pt src="',
      _eProjectsPath,projectName,'.js">',
      '</scri','pt>'].join("")); }
