/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- uiChk ------------------------- 20190927


*** ui checkbox module


The ui checkbox module extends ui base with checkbox ui element.

Implementation note: vs2 overlaps/is duplicates of uiExt module and
uwill overwrite each other. Therefore - on changes - changes have to
be applied to both of them to maintain consistent outcome. It is
duplicated in the ui checkbox so that ui extensions does not need to
be loaded when not used. It saves memory / vars with simple UIs.


--- Enable ui base with uiChk and create a checkbox with callback on untouch:

```
var ui = require("ui")       // load ui base module
 　　  .add(require("uiChk")) // add module into base and remove from cache
      ;
```


--- Create - specify - check box

```
// flgs clazz  id    x  y  w s b f  value object /. callback on 'untouch'
ui.c( 3,"chk","c1",140,45,25,0,2,7,"H"
                  ,function(id,v){ (v) ? LED2.set() : LED2.reset() }
                  ,[12,7,3,5,"On"]);
//                  fv t x y  label text
```

Creates, adds to ui, conditionally displays and returns an active(2),
visible(1) =(3) checkbox ("chk") with id "c01". Checkbox is positioned
at 145 @ 40 (left @ top, x @ y) and sized 25 square (both width and
height), has 2(blue) / 7(white) border / fill colors, is initially
unchecked (0, but any falsey works), has value object "H" (when checked
and undefined when unchecked), has callback that sets and resets green
LED2 accordingly, and is labeled "On" right next to it in fontVector
(size) 12, text (font) color 7(white) on black screen background color
and x / y-offset of 3 / 5 from left top corner of checkbox's bounding box.

Colors are bit-coded with 3-bit color-depth according ui.clrs=[...] setup.

For ui base, color, label, and callback details see (also) ui base module.

Callback cb is called on untouch with ```id, v, ui, e, t``` as arguments:

```
args[] sym   description 
  0    id  : button id ("c1")
  1    v   : value object ("H" - can be any object);
             NOTE: is undefined when check box is unchecked
  2    ui  : ui (THE ui object)
  3    e   : uiChk element (chk 'object' (runtime data structure))
  4    t   : touch info x, y,...  (where last touched)
        { x : x coordinate
        , y : y coordinate
        , t : touched (truey) / untouched (falsey)
        , f : flags
        }
```

For detailed ui setup, including color depth and custom colors, connecting
to display and touch screen, soft key board or buttons, see ui module and
example material (in the ui.zip file and sources with comments). Take a
look at the example that shows all ui elements on one single display -
```uiExampleAll.js``` - project file in the _sbx Espruino sandbox projects
folder. Make it the Espruino Web IDE sandbox folder and run the ready-made
examples on your Espruino board. Espruino forum has several entries about
the modules and hardware. Search in (google) for:
```Espruino Forum allObjects touch screen display modular ui framework```

No board or display with touch screen yet? No Problem: Open ```uiExample.html```
in uidev folder in Web browser to run the same example in the Cross 
Development / Emulation environment (where modules are developed and
maintained).

For helpful details to use the ui base and ui element APIs, take a look
at documentation in ui base and uiBtn modules.


--- chk ui element constructor arguments (a[]) and runtime data structure (e[]) are:

```
arg runtime 'object' instance of 'clazz' chk (chkbox)
a[]  e[]
 0   [0] f  - flags focus(4), active(2), visible(1)
 .    .         0bxx1 visible &1 visible 
 .    .         0bx1x active  &2 active / senses touches vs read/display-only
 .    .         0b1xx focus   &4 focus by touch down, drag w/in bounding box
 1   [1] c  - clazz "chk"
 2   [2] i  - id eg "c01", short, at least 2..3 chars,  ui globally unique.
              Single letter ui element ids are 'reserved' (for keyboard(s)).
 3   [3] x  - x ((left ) of focus / touch bounding box)
 4   [4] y  - y ((top  ) of focus / touch bounding box)
 5       w  - width and height (of focus / touch bounding box)
     [5] x2 - x ((right) of focus / touch bounding box: x - w + 1)
 6       s  - initial checked(truey/1)/unchecked(falsy/0) state used
              with [9] value info
     [6] y2 - y ((bot  ) of focus / touch bounding box: y - h + 1)
 7   [7] bc - border color
 8   [8] fc - fill color
 9       v  - value (returned value when checked)
     [9] vi - value info array w/ state and value
         [0] - truey/falsey for checked/unchecked state of checkbox
         [1] - value object (returned when checkbox is checked)
10  [10] cb - simple, preferred callback on untouch after touchdown
11  [11] l  - label (info), array with:
      l[0]  fv - fontVector (size)
      l[1]  tc - (label) text color
      l[2]  x  - x offset from focus box x ( bounding box left )
      l[3]  y  - y offset from focus box y ( bounding box top  )
      l[4]  tx - label text to display (using .drawString())
12  [12] ca - NON-preferred, experimental callback on any touch event
```

--- chk - checkbox ui element properties - variables and methods - mixed into ui base:

 */
exports = // "chk" (checkbox) 'clazz' name
{ mn: "uiChk" // module name - globally unique (used to remove code from cache)
, chkC:    function (f, c, i, x, y, w,     s,     bc, fc, v, cb,  l, ca) { // constructor
// ----------> chk e[0  1  2  3  4  5(x2)  6(s)    7   8  9  10  11  12] - same as args
    return this.rdc([f, c, i, x, y, x+w-1, y+w-1, bc,fc,[!!s,v],cb,l,ca]); } // runtime obj
, vs2: function(x,y,x2,y2,p) { // vertices for for chk like shapes ('beveled') corners
    // 4 'beveled corners' = 8 corners defined by 0 and p insetting combinations for x, y
    return [x,y+p, x+p,y, x2-p,y, x2,y+p, x2,y2-p, x2-p,y2, x+p,y2, x,y2-p]; }
, chkG: function(e) { return (e[9][0]) ? e[9][1] : undefined; } // get chk VALUE (!state) 
, chkU: function(_,e,t,s,p) { var v=!!s; if (e[9][0]!==v) {
      e[9][0]=v; if ((e[0]&1) && _.di) { _.chkDu(_,e[3],e[4],e[5],e[6],e[7],e[8],v,1); } }
      if (p=(p&&(p=e[10]))?p:e[12]) { if (p(e[2],_.chkG(e),_,e,t)) { _.d(e); } } }
, chkD: function(f,c,i,x,y,x2,y2,bc,fc,vi,cb,l) { var m=2, p=3, _=this;
    _.clr(bc).dsp.fillPoly(_.vs2(x+m,y+m,x2-m,y2-m,p));
    _.chkDu(_,x,y,x2,y2,bc,fc,vi[0],fc!==bc);
    if (l) { _.ld(x2,y,l); } }
, chkDu: function(_,x,y,x2,y2,bc,fc,s,b) { var m=5, p=2, f=2, i=3, c;
    if (b) {_.clr(fc).dsp.fillPoly(_.vs2(x+m,y+m,x2-m,y2-m,p)); }
    if (s) { if (fc===bc) { _.clr(fc,1); } else { _.clr(bc); } c=i+f*2; f=i+f;
      _.dsp.fillPoly([x+f,y+c , x2-c,y2-f, x2-f,y2-c, x+c,y+f ]);
      _.dsp.fillPoly([x+f,y2-c, x2-c,y+f , x2-f,y+c , x+c,y2-f]); } }
, chk: function(_,e,t,c) { // console.log(".chk",c,e[2],(_.ef)?_.ef[2]:"..",(_.lf)?_.lf[2]:"..",t);
    _.chkU(_,e,t,(c)?!e[9][0]:e[9][0],c); } // is in focus untouch (flip) - else any elt and touch event
};
