/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- uiBtn ------------------------- 91 vars (349 w/ ui)

ui button module provides button functionality to ui base module.

--- Enable ui with uiBtn element functions:

```
var ui = require("ui")        // load ui base module
 　　  .add(require("uiBtn")); // add uiBtn module into base and remove from cache
```  

--- Create and add to ui an active(2), visible(1) =(3) button (btn) with id "bt01" 
Button is positioned at 10 @ 20 (left@top x@y) and sized 50 x 30 (width x height),
has 4(red)/7(white) border/fill colors, value object string "B_1" and (arguments
ignoring) callback that toggles red LED1, and is labeled "RED" in fontVector
(size) 20, font color 0(black) and x/y-offset 5/15.

Colors are bit-coded with 1-bit color-depth according ui.clrs=[...] setup.

```
ui.c(3,"btn","bt01",10,50,30,4,7,"B_1"
                   ,function(){ (LED1.read()) ? LED1.reset() : LED1.set() }
                   ,[20,0,5,15,"RED"]);
```

Callback is called on release (untouch) and provides id, v, ui, e, t as arguments:
args[] sym   description 
  0    id  : button id ("bt01")
  1    v   : value object ("B_1"), can be any object
  2    ui  : ui (THE ui object)
  3    e   : uiBtn element (btn 'object' (runtime data structure) just released)
  4    t   : touch info (x, y,...; to calc where within btn btn was 'hit' & released

For detailed ui setup, including color depth and custom colors, connection to
display and touch screen or buttons, see ui module.


// btn - Button ui element methods (like prototype to mix into ui base)
//
// 'constructor' called by ui.c([f,"btn","btnId01",...] with ui[args[1]+C](args);
//
// btn ui element (constructor arguments and runtime data structure):
// arg runtime 'object' instance of 'clazz' button
// a[] e[]
//  0  [0] flags focus(4), active(2), visible(1)
//  .   .  0bxxxxx1 visible    &1 visible 
//  .   .  0bxxxx1x active     &2 active / senses touch down (vs. read/display-only)
//  .   .  0bxxx1xx focus      &4 focus (by touchdown or drag in/over)  
//  .   .  0bxx1xxx touchdown  &8 got touchdown
//  1  [1] class - "btn"
//  2  [2] id - eg "b01", short, null or at least 3 chars, and ui globally unique
//  3  [3] x  - x ((left ) of focus / touch bounding box)
//  4  [4] y  - y ((top  ) of focus / touch bounding box)
//  5      w  - width (of focus / touch box,... 
//     [5] x2 - x ((right) of focus / touch bounding box: x - w + 1)
//  6      h  - height (of focus / touch box,...
//     [6] y2 - y ((bot  ) of focus / touch bounding box: y - h + 1)
//  7  [7] bc - border color
//  8  [8] fc - fill color
//  9  [9] v  - value - any object, from simple string to number to complex { } object
// 10  [10] cb - callback on untouch after touchdown w/ args (id, null, ui, e/btn, t/event)
// 11  [11] l  - label (info), array with:
  // 0   l[0]  fv - fontVector (size)
  // 1   l[1]  tc - (label) text color (index)
  // 2   l[2]  x  : x offset from focus box x ( bounding box left )
  // 3   l[3]  y  : y offset from focus box y ( bounding box top  )
  // 4   l:4]  tx : label text to display (using .drawString())
//

 */
/*
exports=
{mn:"uiBtn"
,btnC:function(b,c,a,f,e,d,g,h,k,l,m){
  return this.rdc([0,b,c,a,0,f,e,f+d-1,e+g-1,h,k,l,m])}
,vs3:function(b,c,a,f,e,d){
  return[b,c+d,b+e,c+e,b+d,c,a-d,c,a-e,c+e,a,c+d,a,f-d,a-e,f-e,a-d,f,b+d,f,b+e,f-e,b,f-d]}
,btnD:function(b,c,a,f,e,d,g,h,k,l,m,p,n){
  b=Math.ceil((Math.min(h-d,k-g)-1)/20);c=3*b;a=1;
  l!==m&&(this.clr(l).dsp.fillPoly(this.vs3(d+a,g+a,h-a,k-a,b,c)),a=4,b=Math.ceil(b-b/a),c=3*b);
  this.clr(m).dsp.fillPoly(this.vs3(d+a,g+a,h-a,k-a,b,c));
  n&&this.ld(d,g,n)}
,btn:function(b,c,a){
  b.fh(b,c,a)&&!a.t&&(b.blr(c),c[11]&&c[11](c[2],void 0,b,c,a)&&b.d(c))}
  if (_.fh(_,e,t) && !t.t) { _.blr(e); if (e[10]) { if (e[10](e[2],e[9],_,e,t)) { _.d(e)) } } }
};
 */
exports = // btn (button)
{ mn: "uiBtn" // module 'clazz' name - globally unique (used to remove code from cache)
, btnC:    function (f, c, i, x, y, w,     h,     bc, fc,  v, cb,  l) { // constructor
// ----------> btn e[0  1  2  3  4  5(x2)  6(x2)   7   8   9  10  11] - same as arguments
    return this.rdc([f, c, i, x, y, x+w-1, y+h-1, bc, fc,  v, cb,  l]); } // runtime obj
, vs3: function(x,y,x2,y2,p,q) { // return vertices for btn like shapes ('round' corners)
    // 12 'round corners' (4x3) defined by 0, p and q insetting combination for x and y 
    return [x,y+q, x+p,y+p, x+q,y,  x2-q,y, x2-p,y+p, x2,y+q,
          x2,y2-q, x2-p,y2-p, x2-q,y2, x+q,y2 ,x+p,y2-p ,x,y2-q]; }
, btnD: function(f,c,i,x,y,x2,y2,bc,fc,v,cb,l) { // D(isplay of) btn (ui element incl. label)
    if (f&1) { var _=this, p=Math.ceil((Math.min(x2-x,y2-y)-3)/20), q=p*3, m=2;
      if (bc!==fc) { _.clr(bc).dsp.fillPoly(_.vs3(x+m,y+m,x2-m,y2-m,p,q)); 
        m=5; p=Math.ceil(p-p/m); q=p*3; }
      _.clr(fc).dsp.fillPoly(_.vs3(x+m,y+m,x2-m,y2-m,p,q));
      if (l) { _.ld(x,y,l); } } }
, btn: function(_,e,t) { // touch event on btn ui element
    console.log("btn():...");
    var f = _.fh(_,e,t);
    console.log("btn(): f="+f+" - t.t="+t.t+" - f'="+(e[0]&4));
    if (f && !t.t) { _.blr(e); // touched down and had focus and is now released
      if (e[10] && e[10](e[2], e[9], _, e, t)) { _.d(e); }; }
    console.log("btn().");
  }
};
