/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- uiInp ------------------------- 144 vars

ui input module provides input field functionality to ui base module.

--- Enable ui with uiInp element functions:

Note: requires uiExt(ension) module (see uiExt.js module)

```
var ui = require("ui")       // load ui base module
      .add(require("iuExt")) // add uiExt module into base and remove from cache
      .add(require("uiInp")) // add uiInp module into base and remove from cache
      ;
```  

--- Create and add to ui an active(2), visible(1) =(3) input field at x/y (left/top:
45/7) of size width/height (195/28), turkis border color (3), white

Colors are bit-coded with 1-bit color-depth according ui.clrs=[...] setup.

```
ui.c(3,"inp","i1",45, 7, 195, 28, 3, 7, "modular UI", cb,[[13, 0,   5, 6, 16,   0,  0]
                                                         ,[13, 7,-233, 6, "Field"    ]
                                                         ]);
```

Callback cb is called on release (untouch) and change and provides
```id, v, ui, e, t``` as arguments:

```
args[] sym   description 
  0    id  : button id ("ck01")
  1    v   : value object ("H" - can be any object - and undefined when unchecked)
  2    ui  : ui (THE ui object)
  3    e   : uiInp element (inp 'object' (runtime data structure) just released)
  4    t   : touch info x, y,... object to calc where within chk chk was 'hit' & released
        { x : x coordinate
        , y : y coordinate
        , t : touched (truey) / untouched (falsy)
        }
```

For detailed ui setup, including color depth and custom colors, connection to
display and touch screen or buttons, see ui module.

'constructor' is called by ```ui.c(3,"rad","radId01",...)``` with ```ui[args[1]+C](args);```
and 'creates' and registers rad ui element, conditionally displays it, and returns it.

inp ui element constructor arguments and runtime data structure are:

```
arg runtime 'object' instance of 'clazz' button
a[]  e[]
 0   [0] flags focus(4), active(2), visible(1)
 .    .  0bxx1 visible  &1 visible 
 .    .  0bx1x active   &2 active / senses touch down (vs. read/display-only)
 .    .  0b1xx focus    &4 focus (by touchdown or drag in/over)  
 1   [1] clazz - "inp"
 2   [2] id - eg "i01", short, null or at least 3 chars, and ui globally unique
              Single letter ui element id's are 'reserved' (for keyboard(s)).
 3   [3] x  - x ((left ) of focus / touch bounding box)
 4   [4] y  - y ((top  ) of focus / touch bounding box)
 5       w  - width (of focus / touch box,... 
     [5] x2 - x ((right) of focus / touch bounding box: x - w + 1)
 6       h  - height (of focus / touch box,...
     [6] y2 - y ((bot  ) of focus / touch bounding box: y - h + 1)
 7   [7] bc - border color
 8   [8] fc - fill color
 9   [9] v  - value - (string) value to start with
10  [10] cb - callback on untouch after touchdown w/ args (id, value, ui, e/btn, t/event)
11  [11] ls - labels - array of labels of which the first one is also value related   
      ls[0] label (info) for value related info, array with:
         l[0]  fv - fontVector (size)
         l[1]  tc - (label) text color (index)
         l[2]  x     - x offset from focus box x ( bounding box left )
         l[3]  y     - y offset from focus box y ( bounding box top  )
         l[4]  mxLen - maximum length
         l[5]  type  - optional type (absent and 0 being plain string and only for now)
         l[6]  frmt  - optional formatter function to use to format value for label
      ls[1,2,3,...] any number of additional labels, but mostly just two for min and max-
         l[0]  fv - fontVector (size)
         l[1]  tc - (label) text color (index)
         l[2]  x  - x offset from focus box x ( bounding box left )
         l[3]  y  - y offset from focus box y ( bounding box top  )
         l[4]  tx - label text to display (using .drawString())
```

inp - input field ui element methods (like prototype to mix into ui base):

 */
exports = // inp (input field)
{ mn: "uiInp" // module 'clazz' name - globally unique (used to remove code from cache)
, inpC:     function(f, c, i, x, y, w,     h,     bc, fc,  v,  cb, ls) {
// ----------> btn e[0  1  2  3  4  5(x2)  6(y2)   7   8   9   10  11] - same as arguments
    return this.rdc([f, c, i, x, y, x+w-1, y+h-1, bc, fc,  v,  cb, ls]); }
, inpG: function(e) { return e[9]; } // G(et value from) inp(ut ui element)
, inpU: function(_,e,t,v,p) { var vi=e[11][0], v2=("string" === typeof v) // update
    ? v : ""+v, l; v2=v2.substr(0,Math.min(v2.length, (l=vi[4]) ? l : v2.length));
    if (v2!==e[9]) { e[9]=v2; _.inpDu(_,e[0],e[3],e[4],e[5],e[6],e[8],v2,e[11][0],1); }
    if (p && e[10] && e[10](e[2], v2, _, e, t)) { _.d(e); } }
, inpD: function(f,c,i,x,y,x2,y2,bc,fc,v,cb,ls) { // D(isplay of) inp(ut ui element)
    // displays whole ui element incl. label; uses inpDu to display the updateable part
    var _=this, m=2; _.clr(bc).dsp.fillPoly(_.vs3(x+m,y+m,x2-m,y2-m,1,3));
    _.inpDu(_,f,x,y,x2,y2,fc,v,ls[0],bc!==fc);
    if (m=ls[1]) { _.ld(x2,y,m); } }
, inpDu: function(_,f,x,y,x2,y2,fc,v,vi,b) { // Display (redraws) the updateable part of inp
    var m=5; if (b) { _.clr(fc).dsp.fillRect(x+m,y+m,x2-m,y2-m); } if (f&2) {
      _.clr(fc,1).dsp.drawLine(x2-11,y+11,x2-4,y+4);_.dsp.drawLine(x2-11,y+4,x2-4,y+11); }
    _.ld(x,y,[vi[0],vi[1],vi[2],vi[3],(m=vi[7])?m(v):v]); }
, inp: function(_,e,t) { if ((e===_.ef) && !t.t) { // touch event on inp ui element
      _.inpU(_, e, t, ((Math.abs(_.lx-_.dx)<4) && (Math.abs(_.ly-_.dy)<4)
        && (Math.abs(e[5]-7-t.x)<4) && (Math.abs(e[4]+7-t.y)<4)) ? "" : e[9], 1); } }
};      

if (Modules._eCache) { Modules._eCache(exports.mn, exports); } // for development in emulator only
