/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- soft / touch keyboard / keypad ------------------------- 20160101 - 141 vars


*** soft / touch keyboard / keypad 


uiKbd3x10B - ui keyboard module extends ui base with full keyboard ui element.

ui keyboard module is built with 3 rows x 10 columns of uiBtn ui elements and
therefore requires uiBtn ui element to be loaded.

Keyboard ey ids are the button ids, and button ids are  SINGLE character ids
(thus reserved for the ui modules and not available to application UI definitions.
The id of a key is the 1st char on the key (btn), for examples:

```
 - "a" for key (btn) "a@"  , returning the values "a"    , "A", and "@".
 - "<" for key (btn) "<' \", returning the values "cr/lf", "'", and "\".
 - "." for key (btn) "
```

Keyboard key handler calls cb callback w/ values v, k, u, e, and t parameters:

```
  v - key value - single character representing the printable key; special values:
      "sh0".."sh3" for shift key, "bs: for BacksSpace, "cr/lf" for CR/LF  
  k - keyboard object (this)
  u - ui core singleton
  e - event resoures
  t - touch object
  f - flags
```

e is the key board button that has just been untouched (k: keyboard, u: ui core
and t: touch event object).

Key value depends on short TAP / LONG TOUCH and shift key mode.

^-shift key cycles w/ short taps through  ^ m=0,  ^^ m=1,  ^^^ m=2,  ^# m=3 its modes;
  long touch always goes directly to  ^# m=3 shift mode (number and special characters).

For REGULAR keys, shift key modes and returned chars with regular - short tap =
are (long touch always returns 3rd character on key if there is, else none):

```
  ^    0: lower case   - key's id / 1st char to lowercase
  ^^   1: 1 upper case - key's id / 1st char to uppercase and falls back to mode 0.
  ^^^  2: all upper case, key's id / 1st char to uppercase
  ^#   3: all number and special chars - 2nd char on key
```

NOTE2: ^shift key also returns a key value: "sh"+mode: "sh0", "sh1", "sh2", "sh3".

Special keys with shift mode 0..4 returns (returns listed by key and mode):

```
  spec key mode m    returns  (comment)
  < cr/lf  
           ^    0:   cr/lf    ("cr/lf" string)
           ^^   1:   '        (single quote)
           ^^^  2:   \        (back slash)
           #^   3:   cr/lf    ("cr/lf" string)
  . dot
           ^    0:   ?        (question mark)
           ^^   1:   .        (dot)
           ^^^  2:   .        (dot) 
           #^   3:   .        (dot)
  b/blank... 
    ... w/ short tap:
           ^    0:   b/blank  (space 'bar')
           ^^   1:   "        (double quotes)
           ^^^  2:   _        (underscore)
           #^   3:   b/blank  (space 'bar')
    ... w/ long touch:
           N/A       bs       ("bs" string for BackSpace)
```

Usage example:

get the keyboard and 'connect' it to entry field "i1" - as part of UI definition:

Note: this keyboard modules requires ui base and uiBtn modules to be loaded. It creates
for each key a button with the id being the first char displayed on the key, for example,
a,b,c,..., ^ for shift, < for enter/return, blank (" ") for blank/backspace,... Therefore,
(practically) all single letter ui element id's are taken and therefore 'reserved'.

```
var ui = require("ui")       // load ui base module
 　　  .add(require("uiBtn")) // add uiBtn module into base and remove from cache
      ;
{"uiKbd3x10B"}.connect(                 ui,x,  y, w, h,bc,fc,sc,mx,my,fs,tc,tx,ty,cb) {
var kbd = require("uiKbd3x10B").connect(ui,0,215,24,24, 7, 7,-2, 1, 1,10 ,0, 2, 4,
  function(v){
    var s = ui.g("i1"); // is uiInp(ut field) ui element 
    if (v.length === 1) {  'normal' key - append
      ui.s("i1",s + "" + v);
    } else if (v === "bs") { BS/BackSpace key - remove last char
      if (s.length > 0) { ui.s("i1",s.substr(0,s.length - 1)); } 
    }
  });
```

.connect(arguments) are:

```
 0  ui  ui core
 1  x   top bounding box border
 2  y   left bounding box border
 3  w   width of bounding box
 4  h   height of bounding box
 5  bc  key border color
 6  fc  key fill color
 7  sc  special key fill color for special keys
 8  mx  margin of key bounding box
 9  my  margin of 
10  fv  key fontVector (size)
11  tc  key text color 
12  tx  key text x offset (within btn bounding box)
13  ty  key text y offset (within btn bounding box)
14  cb  callback function to accept following parameters:
   0  v   key: a single char, "bs" for backspace, "cr/lf", "sh0".."sh4" for shift key
   1  k   keyboard (this sigleton)
   1  u   ui core singleton (see ui module)
   2  e   key btn that was just released (see uiBtn module)
   3  t   touch object (see ui module)
```
 */
exports =  // uiKbd3x10B
{ mn: "uiKbd3x10B"
, lt: 0.3  // long touch min time in seconds
, tt: 0    // touch time in secs
, sm: 0    // shift mode (shift key text and mode: "^"=0, "^^"=1, "^#"=2, "^^^"=3)
, fk: null // first key ("q1`")
, sk: null // shift key
, cb: null // callback
, st: ["^","^^","^^^","#^"]
, connect:   function(ui,x,  y, w, h,bc,fc,sc,mx,my,fs,tc,tx,ty,cb) {
    // parms example: ui,0,215,24,24, 7, 7,-2, 1, 1,10, 0, 2, 4,
    var _=this, f = (function(id,d,u,e,t) { // key handler (key btn callback)
          var m=_.sm, v=null, s=((_.tt=getTime() - u.tt) < _.lt);
          if ((p="^<. ".indexOf(id))<0) { // regular keys (!^shifts && !cr/lf && !. && !b(lank))
            if (s) { // short tap
              if (m==1) { // ==1:1UC 
                v = (id == ".") ? "/" : id.toUpperCase(); m=0;
              } else { // 0:*lc | 2:*UC: | 3:*sp:  
                v = (m) ? ((m == 2) ? id.toUpperCase() : e[11][4].charAt(1)) : id; }
            } else { // long touch - 3rd char
              if ((p = e[11][4]).length > 2) { v = p.charAt(2); } m = (m == 1) ? 0 : m; }
          } else if (!p) { // (p==0) (id==="^") { // ^ - shifts
            v = "sh" + ( m = (s) ? ((m = m + 1) > 3 ? 0 : m) : 3 ); // short tap | long touch
          } else if (p == 1) { // (id==="<") { // < CR/LF: 0:*lc: cr/lf | 1:1UC: special ' | 2:*UC: \ | 3:*sp: cr/lf
            if (m == 1) {  v = "'"; m = 0; } else { v = (m == 2) ? "\\" : "cr/lf"; }
          } else if (p == 2) { // (id===".") { // .: 0=*lc: . | 1=1UC: special: ? | 2=*UC: . | 3=*sp: .
            if (m == 1) { v = "?"; m = 0; } else { v = "."; }
          } else { // (p===3) (id===" ") { // b/blank: 0=l*c: b | 1=1UC: special: " | 2=*UC: b | 3=*sp: b
            if (s) { if (m == 1) { v = "\""; m = 0; } else { v = (m==2) ? "_" : " "; } // short tap
            } else { v = "bs"; } } // long touch - BackSpace
          if (m!==_.sm) { _.sk[11][4]=_.st[_.sm=m]; _.cb(v,k,u,e,t); if (u.di) { u.d(_.sk); }
          } else if (v !== null) { _.cb(v,k,u,e,t); }
        }).bind(_),
        xr, yr=y-h+my, k="q1w2e3r4t5y6u7i8o9p0a@s#d$f%g^h&j~k(l)<'^ z-x+c=v<b>n!m,.? \"",
        es=ui.es, r=es.length, bw=w-mx-mx, bh=h-my-my, i;
    // creating the basic key with letter and numeric/special character
    for (i=0; i<59; i+=2) { if (i%20 === 0) { xr=1+mx; yr+=h; } else { xr+=w; } 
      ui.c(3,"btn",k.charAt(i),xr,yr,bw,bh,bc,fc,undefined,f,[fs,tc,tx,ty,k.substr(i,2)]); }
    // color the special keys: enter/shift/dot/space 
    [19,20,28,29].forEach(function(rr){ (k=es[r+rr])[7]=k[8]=sc; });
    // add more types to certain keys (including blanks for 'formatting'
    xr="`;:*[] \\{}/|.. _";
    [0,14,15,16,17,18,19,19,24,25,26,27,28,28,29,29].forEach(function(rr,i){ 
        yr=(k=es[r+rr])[11]; yr[4]+=xr.charAt(i); if (ui.di) { ui.d(k); } });
    _.fk=es[r]; _.sk=es[r+20]; _.cb=cb;
    return _;
  }
};
