/* Copyright (c) allObjects - See the file https://muet.mit-license.org LICENSE for copying permission. */
/*

----- uiSli ------------------------- 135 vars

ui slider module provides slider functionality to ui base module.

Setup ui control with uiSli element functions:

```
var ui = require("ui")       // load ui base module
      .add(require("uiSli")) // add uiSli module into base and remove from cache
      ;
```

Create and add a 0..100% slider to the UI that shows % range (min %:0 and max 100) as
text labels left and right of the slider and current value - on untouch / set - in 
center on slider. Value label is formatted with provided custom formatter function:

```
ui.c(3,"sli","s2" , 30,175,180, 25,  3,  3,  33, cb, [[12, 6, 70, 5, 0, 100
                                                ,function(v,_){       // (value, ui)
                   if (_.it) { return ""; } // show value label only on untouch / set
                   var s=Math.round(v*10)+"%"; return (((v<1)?"0":"") // labels:
                       +s.substr(0,s.length-2)+"."+s.substr(-2)); }]  // val % on slider
                                                ,[12, 7,-25,5,"%:0"]  // min to the left
                                                ,[12, 7,182,5,"100"]  // max to the right
                                                     ]);
```

Callback (cb) is called on touch, change, release (untouch) and provides 
```id, v, ui, e, t``` as arguments:

```
args[] sym   description 
  0    id  : button id ("s2")
  1    v   : value object 44.35262224
  2    ui  : ui (THE ui object)
  3    e   : uiSli element (sli 'object' (runtime data structure))
  4    t   : touch info x, y,... object to calc where within btn btn was 'hit' & released
        { x : x coordinate
        , y : y coordinate
        , t : touched (truey) / untouched (falsy)
        }
```

For detailed ui setup, including color depth and custom colors, connection to
display and touch screen or buttons, see ui module.

'constructor' is called by ```ui.c(f,"sli","s02",...)``` with ```ui[args[1]+C](args);```
and 'creates' and registers sli(der) ui element, conditionally displays it, and returns it.

sli(der) ui element constructor arguments and runtime data structures are:

```
arg runtime 'object' instance of 'clazz' slider
a[]  e[]
 0   [0] flags focus(4), active(2), visible(1)
 .    .  0bxx1 visible  &1 visible 
 .    .  0bx1x active   &2 active / senses touch down (vs. read/display-only)
 .    .  0b1xx focus    &4 focus (by touchdown or drag in/over)  
 1   [1] class - "sli"
 2   [2] id - eg "s02", short, null or at least 3 chars, and ui globally unique
              Single letter ui element id's are 'reserved' (for keyboard(s)).
 3   [3] x  - x ((left ) of focus / touch bounding box)
 4   [4] y  - y ((top  ) of focus / touch bounding box)
 5       w  - width (of focus / touch box,... 
     [5] x2 - x ((right) of focus / touch bounding box: x - w + 1)
 6       h  - height (of focus / touch box,...
     [6] y2 - y ((bot  ) of focus / touch bounding box: y - h + 1)
 7   [7] bc - border color
 8   [8] fc - fill color
 9   [9] v  - value - a number
10  [10] cb - callback on touch, value change, untouch w/ args (id, value, ui, e/btn, t/event)
11  [11] ls - labels - array of labels of which the first one is also value related   
      ls[0] label (info) for value related info, array with:
         l[0]  fv - fontVector (size)
         l[1]  tc - (label) text color (index)
         l[2]  x  - x offset from focus box x ( bounding box left )
         l[3]  y  - y offset from focus box y ( bounding box top  )
         l[4]  mn - minimum value ( at left  border )
         l[5]  mx - maximum value ( at right border )
         l[6]  fm - optional formatter function to use to format value for label
      ls[1,2,3,...] any number of additional labels, but mostly just two for min and max-
         l[0]  fv - fontVector (size)
         l[1]  tc - (label) text color (index)
         l[2]  x  - x offset from focus box x ( bounding box left )
         l[3]  y  - y offset from focus box y ( bounding box top  )
         l[4]  tx - label text to display (using .drawString())
```

sli - Slider ui element methods (like prototype to mix into ui base):
 */
exports = // sli (slider)
{ mn: "uiSli" // module 'clazz' name - globally unique (used to remove code from cache)
, sliC:     function(f, c, i, x, y, w,     h,     bc, fc,  v,  cb, ls) {
// ----------> sli e[0  1  2  3  4  5(x2)  6(y2)   7   8   9   10  11] - same as arguments
    return this.rdc([f, c, i, x, y, x+w-1, y+h-1, bc, fc,  v,  cb, ls]); }
, sliG: function(e) { return e[9]; } // get slider value 
, sliU: function(_,e,t,v,p) { // update slider value w/ or w/o p(ropagation) (call back)
    var l=e[11][0]; e[9]=v; if (p) { if (e[10]&&e[10](e[2],v,_,e,t)) { _.d(e); return; } }
    _.sliDu(_,e[3],e[4],e[5],e[6],e[7],e[8],v,l,l[4],l[5],1); }
, sliD: function(f,c,i,x,y,x2,y2,bc,fc,v,cb,ls) { // display slider
    var _=this, l=ls[0], lx=ls.length; _.clr(bc).dsp.fillRect(x+4,y+5,x2-4,y2-5);
    _.sliDu(_,x,y,x2,y2,bc,fc,v,l,l[4],l[5]); while (--lx>0) { _.ld(x,y,ls[lx]); } }
, sliDu: function(_,x,y,x2,y2,bc,fc,v,l,mn,mx,m) { // update slider display
    var w=x2-x-9, w1=Math.round(w/(mx-mn)*(v-mn));
    if (m && w1) { _.clr(bc).dsp.fillRect(x+5,y+6,x+5+w1-1,y2-6); }
    _.clr((fc!==bc)?fc:_.bc).dsp.fillRect(x+5+w1,y+6,x2-5,y2-6);
    if (w=l[0]) { _.ld(x,y,[w,l[1],l[2],l[3],(w1=l[6])?w1(v,_):v]); } }
, sli: function(_,e,t) { var x=e[3], x2=e[5], l=e[11][0], v=e[9], mn=l[4], mx=l[5], d=t.t,
    w=x2-x-10, vw=mx-mn, w1=Math.round(w/vw*(v-mn)), w3=t.x-(x+5)+1, wd=w3-w1, v2;
    if ((Math.abs(wd)>3) || (wd&&((w3<3) || (w3>(w-3)))) || !d) { w3=(w3<0)?0:(w3>w)?w:w3;
      if (((v2=mn+w3/w*vw)!=v)||!d) { _.sliU(_,e,t,v2,d); } } }
};
