// TFT24TRCportraitCalibr3.js

// 20190919 - allObjects - variables:  /  / 

//  2.4" 240x320 262K Color TFT LCD ILI9341 on SPI1 with 
//  Resistive Touch Screen Controller XPT2046 on SPI2
//  using out of box ILI9341 and ADS7843 modules.
//
// Wiring:
//  2.4"
//  DSP                  PICO [<---USB
//   1   VCC      red     3   3.3            shared
//   2   GND      blk     1   GND            shared
//   3   CS       blu     7   B6   CS        LCD
//   4   RST      brn     9   A8   reset RST LCD
//   5   D/C      grn     8   B7   D/C       LCD
//   6   MOSI     ylw     6   B5   SPI1 MOSI LCD
//   7   SCK      wht     4   B3   SPI1 SCK  LCD
//   8   LED      org    18   A5   LED       LCD
//   9   MISO     grn     5   B4   SPI1 MISO LCD
//  10   T_CLK    wht    23   B13  SPI2 SCK  Touch
//  11   T_CS     ylw    22   B10  CS        Touch
//  12   T_MOSI   grn    25   B15  SPI2 MOSI Touch
//  13   T_MISO   blu    24   B14  SPI2 MISO Touch
//  14   T_IRQ    blk    21   B1   IRQ       Touch

var dW = 180; // width  240 pixel
var dH = 260; // height 320 pixel

var dsp, dMod = require("ILI9341"),
    tSPI=SPI2, tCs = B10, tIrq = B1, 
    touch, tMod = require("XPT2046");

// marker specs (lik dice face 5; top left to bottom right)
var mrks =
    [{i:1, label:"topLeft    ", x:   20, y:   20, xOff: 14, yOff:-6}
    ,{i:2, label:"topRight   ", x:dW-20, y:   20, xOff:-66, yOff:-6}
    ,{i:3, label:"center     ", x:dW/2 , y:dH/2 , xOff:-29, yOff:14}
    ,{i:4, label:"bottomLeft ", x:   20, y:dH-20, xOff: 14, yOff:-6}
    ,{i:5, label:"bottomRight", x:dW-20, y:dH-20, xOff:-74, yOff:-6}
    ];
var mSiz = 9; // half of marker size, (0.7 of font size)

// print marker at x/y of size s with x/y label
function mrkr(d,x,y,s,xOff,yOff) {
  dsp.setColor(1,1,1);
  dsp.drawRect(x-s,y-s,x+s,y+s);
  dsp.drawLine(x-s,y,x+s,y);
  dsp.drawLine(x,y-s,x,y+s);
  dsp.setFontVector(s / 0.7);
  dsp.drawString(x + " / " + y, x+xOff, y+yOff);
}

function average(arry) { // return avarage of array values
  return arry.reduce(
    function(p,c) { return p + c; },0) / arry.length; }

function vm(v) { // return value rounded to 1/1000 (millis)
  return Math.round(v * 1000) / 1000;
}

function fInt(i,l) { return ("        " + i).substr(-l); }

function logXY(m,xp,yp,l,d) {
  console.log(
      m.i
    , m.label
    , fInt(m.x,4), fInt(m.y,4)
    , fInt(m[xp],l), (d) ? "(" + fInt(m[xp] - m.x,4) + ")" : ""
    , fInt(m[yp],l), (d) ? "(" + fInt(m[yp] - m.y,4) + ")" : ""
    );
}

function handleAverage(seq,xs,ys) { // log x / y average
  var m = mrks[seq],
      x = Math.round(average(xs)), 
      y = Math.round(average(ys));
  m.xAvg = x; m.yAvg = y;
  logXY(m,"xAvg","yAvg",6);
  if (seq === 4) { // all in, calc calibration parms and calib with new calc function
    console.log("3. ---- calculate calculation function");
    var xscs = [], // xscales = x deltas per pixel
        yscs = [], // yscales = y deltas per pixel
        xoss = [], // xoffsets 
        yoss = []; // yoffsets
    xscs.push((mrks[1].xAvg - mrks[0].xAvg) / (mrks[1].x - mrks[0].x)); // top x deltas
    xscs.push((mrks[2].xAvg - mrks[0].xAvg) / (mrks[2].x - mrks[0].x)); // top left - cent
    xscs.push((mrks[1].xAvg - mrks[2].xAvg) / (mrks[1].x - mrks[2].x)); // cent - top right
    xscs.push((mrks[4].xAvg - mrks[3].xAvg) / (mrks[4].x - mrks[3].x)); // bottom x deltas
    xscs.push((mrks[2].xAvg - mrks[3].xAvg) / (mrks[2].x - mrks[3].x)); // bot left - cent
    xscs.push((mrks[4].xAvg - mrks[2].xAvg) / (mrks[4].x - mrks[2].x)); // cent - top right
    console.log(xscs);
    yscs.push((mrks[3].yAvg - mrks[0].yAvg) / (mrks[3].y - mrks[0].y)); // left y deltas
    yscs.push((mrks[2].yAvg - mrks[0].yAvg) / (mrks[2].y - mrks[0].y)); // top left - cent
    yscs.push((mrks[3].yAvg - mrks[2].yAvg) / (mrks[3].y - mrks[2].y)); // cent - left bottom
    yscs.push((mrks[4].yAvg - mrks[1].yAvg) / (mrks[4].y - mrks[1].y)); // right y deltas
    yscs.push((mrks[2].yAvg - mrks[1].yAvg) / (mrks[2].y - mrks[1].y)); // top right - cent
    yscs.push((mrks[4].yAvg - mrks[2].yAvg) / (mrks[4].y - mrks[2].y)); // cent - bot right
    console.log(yscs);
    var xsc = average(xscs), 
        ysc = average(yscs);
    mrks.forEach(function(m) { 
        xoss.push(m.x - m.xAvg / xsc); 
        yoss.push(m.y - m.yAvg / ysc);
    });
    console.log(xoss);
    console.log(yoss);
    var xos = average(xoss), 
        yos = average(yoss);
    console.log("x / y scale", xsc, ysc);
    console.log("x / y offset", xos, yos);
    mrks.forEach(function(m){
      m.xc = Math.round((m.xAvg / xsc) + xos);
      m.yc = Math.round((m.yAvg / ysc) + yos);
      logXY(m,"xc","yc",4,true);
    });
    
    // calibrated calc function set in touch:
    // x / y scale   -121.5825         89.79285714285
    // x / y offset   258.33405300927 -18.53090446265
    xsc = vm(xsc); xos = vm(xos);
    ysc = vm(ysc); yos = vm(yos);
    console.log("Calciulation function is:");
    console.log("function(yR, xR, d, m) { // portrait 240 x 320");
    console.log("    return [ //rawVal / valPerPx * offset       ");
    console.log("             Math.round(xR / " + xsc + " + " + xos + ")");
    console.log("           , Math.round(yR / " + ysc + " + " + yos + ")");
    console.log("           , d, m");
    console.log("           ];");
    console.log("  }");
    touch.calc = function(yr, xr, d, m) { // xy raw
      return [Math.round(xr / xsc + xos) // / x scale per pixel + x offset
             ,Math.round(yr / ysc + yos) // / y scale per pixel + y offset
             , d, m];
    };
    
    calibrated = true;
    console.log("4. ---- in calibrated, normal mode - validating:");
    console.log("   touch markers for validation.");
  }
}

var calibrated = false;

function onInit() {
  A5.set();
  SPI1.setup({sck:B3, miso:B4, mosi:B5, baud: 1000000});
                   // spi, dc, cs, rst, callback
  dsp = dMod.connect(SPI1, B7, B6, A8, function() {
      dsp.clear();
      // print markers like dice face of 5
      console.log("1. ---- Display markers.");
      mrks.forEach(function(m) {
        mrkr(dsp, m.x, m.y, mSiz, m.xOff, m.yOff); });
      console.log("2. ---- in calibration mode - capturing raw values:");
      console.log("   touch markers from top left to bottom right.");
      // setup touchscreen with callback averaging
      // x and y sequence for calibration of 
      // touchscreen to lcd display.
      SPI2.setup({sck:B13, miso:B14, mosi:B15, baud: 2000000});
      var seq = 0, xs = [], ys = [], x, y;
      touch = tMod.connect(
          // spi, cs, irq, callback, calc // landscape
          SPI2, B10, B1, function(m1, m2, rd) {  // portrait
            var xt, yt;
            if (calibrated) {
              xt = m1; yt = m2;
              if (xt !== undefined) {
                console.log(fInt(xt,4),fInt(yt,4));
              }
            } else {
              xt = m2; yt = m1;
              if (xt !== undefined) {
                // console.log(seq +1,": ",xt,yt);
                // collect x and y for averaging
                xs.push(xt); ys.push(yt);
              } else {
                // log x and y avarages in console 
                handleAverage(seq, xs, ys);
                xs = []; ys = []; seq   = (seq +1) % 5;
              }
            }
      }).listen();
  });
}

onInit();
