// uiX.js - ui and ui elements exploring

// 20190919 - allObjects - variables:  /  / 

//  Modular and extensible UI - with base, ui elements -
//  2.4" 240x320 262K Color TFT LCD ILI9341 on SPI1 with 
//  Resistive Touch Screen Controller XPT2046 on SPI2
//  using out of box ILI9341 and ADS7843 modules
//
// Wiring:
//  2.4"
//  DSP                  PICO [<---USB
//   1   VCC      red     3   3.3            shared
//   2   GND      blk     1   GND            shared
//   3   CS       blu     7   B6   CS        LCD
//   4   RST      brn     9   A8   reset RST LCD
//   5   D/C      grn     8   B7   D/C       LCD
//   6   MOSI     ylw     6   B5   SPI1 MOSI LCD
//   7   SCK      wht     4   B3   SPI1 SCK  LCD
//   8   LED      org    18   A5   LED       LCD
//   9   MISO     grn     5   B4   SPI1 MISO LCD
//  10   T_CLK    wht    23   B13  SPI2 SCK  Touch
//  11   T_CS     ylw    22   B10  CS        Touch
//  12   T_MOSI   grn    25   B15  SPI2 MOSI Touch
//  13   T_MISO   blu    24   B14  SPI2 MISO Touch
//  14   T_IRQ    blk    21   B1   IRQ       Touch

// ----- pull in dispaly and touchscreen
var dW = 240, dH = 320, dsp, dMod = require("ILI9341"),
    touch, tMod = require("XPT2046");

// ----- pull in ui base and ui element
var ui = require("ui")
      .add(require("uiBtn"))
      .add(require("uiSli"))
      .add(require("uiChk"))
      .add(require("uiExt"))
      .add(require("uiRad"))
      .add(require("uiInp"))
      ;

// ----- customize ui colors
ui.clrs = // imply color depth and table with converters and define user colors
  [function(c,i) { // color info to [r,g,b] converter
       var v, s=isNaN(c); if (s || c<0) { // c<0 (looked up [R,G,B]) or c=[R,G,B]
       return (((s) ? c : this.clrs[-c]).map(function(v){ return (i) ? v^255 : v/256; }));
     } else { // c>=0 bit coded rgb color (default 1 bit color depth 2^1^3=8 colors)
       v = (i) ? c^7 : c; return [v>>2&1,v>>1&1,v&1]; } }.bind(ui)     
  ,[216,216,216] // user color -1: light grey // keyboard bottom right special keys
  ,[192,192,192] // user color -2: medium light grey
  ,[128,128,128] // user color -3: medium gray
  ,[192,128,192] // user color -4: light purple
  ,[0  ,160, 64] // user color -5: kind of a 'darker' green
  ];
ui.bc = 0; // ui (screen) background color
ui.di = false; // do not display instantly (on element creation), but deferred


// ----- define generic callback - specific to this development context
var cb =  function(id,v) {
  var l = ("undefined" === typeof log) ? console.log : log;  
  l("ui elt id: " + id + " - value: "
    + ((v===undefined) ? "undefined" : (v===null) ? "null" : v));
};


// ----- define UI (in level 0, so it does not use memory for source)

var u = undefined;


// ---------- Inputfield examples ----------
//   0  1     2       3   4    5    6    7   8  9       10  11 
// flgs clazz id      x   y    w    h   bc  fc  valObj  cb,  ls (arr of label ([[..],..]))
//      btn                 ->x2 ->y2                     [0] fv tc    x  y  mxLen typ fmt
//                                                        [1] fv tc    x  y  label text
ui.c(3,"inp","i1"  , 45,  7, 195,  28,   3,7,"modular UI",u,[[13, 0,   5, 6, 16,   0,  0]
                                                            ,[13, 7,-233, 6, "Field"    ]
                                                            ]);
// ---------- Button examples ----------
//   0  1     2       3   4    5    6    7   8  9       10  11 
// flgs clazz id      x   y    w    h   bc  fc  valObj  cb,  l (label array obj)
//      btn                 ->x2 ->y2                         fv tc    x  y  label text
ui.c(3,"btn","b1"  ,  5, 40,  65,  35,  4,   4, "B_1",  cb,  [15, 7,  13, 9, "RED"  ]);
ui.c(3,"btn","b2"  , 70, 40,  65,  35,  5,   6, {v:1},  cb,  [15, 0,   9, 9, "Y-w-B"]);


// ---------- Checkbox examples ----------
//   0  1     2       3   4    5    6    7   8  9       10  11 
// flgs clazz id      x   y    w    s   bc  fc  valObj  cb,  l (label array obj)
//      chk                 ->x2+->y2~--------->[s,vOjb]      fv tc    x  y  label text
ui.c(3,"chk","c1"  ,140, 45,  25,   0,  -5,  7, "H",    cb,  [12, 7,   3, 6, "On"   ]);
ui.c(3,"chk","c2"  ,192, 45,  25,   1,   1,  1, "X",    cb,  [12, 7,   3, 6, "Up"   ]);


// ---------- Radiobutton examples ----------
//   0  1     2       3   4    5    6    7   8  9       10  11 
// flgs clazz id.grp  x   y    w    s   bc  fc  valObj  cb,  l (label array obj)
//      rad                 ->x2+->y2~--------->[s,vOjb] g    fv tc    x  y  label text
ui.c(3,"rad","r1.a",  5, 80,  37,   0,   6,  2, "L",    cb,  [12, 7,   5,10, "Large"]);
ui.c(3,"rad","r1.b", 95, 85,  27,   1,   3,  1, "M",    0 ,  [12, 7,   5, 7, "Med"  ]);
ui.c(3,"rad","r1.c",170, 90,  17,   1,   5,  2, "S",    0 ,  [12, 7,   5, 2, "Small"]);
ui.c(3,"rad","r2.a",  5,120,  23,   0,   1,  2, "L",    cb,  [12, 7,   5, 4, "Land" ]);
ui.c(3,"rad","r2.b", 95,120,  23,   0,  -3, -3, "W",    0 ,  [12, 7,   5, 4, "Water"]);
ui.c(3,"rad","r2.c",170,120,  23,   1,   1,  1, "A",    0 ,  [12, 7,   5, 4, "Air"  ]);


// ---------- Slider examples ----------
//   0  1     2       3   4    5    6    7   8  9       10  11 
// flgs clazz id      x   y    w    h   bc  fc  valObj  cb, ls (array of label ([[...],..]))
//      sli                 ->x2 ->y2                     [0] fv tc    x  y  min max  format
//                                                         >0 fv tc    x  y  label text
ui.c(3,"sli","s1"  , 10,145, 220,  26,   4, -5, 0,      cb, [[ u, u,   u, u, -10, 10]]);  
ui.c(3,"sli","s2"  , 30,175, 180,  26,  -4, -4, 33,     cb, [[12, 6,  70, 6,   0,100
                                                             , function(v,_){ // (value, ui)
                            if (_.it) { return ""; } // show label on slider only on untouch
                            var s=Math.round(v*10)+"%"; return ((v<1)?"0":"") // labels:
                                 +s.substr(0,s.length-2)+"."+s.substr(-2); }] //   on slider
                                                            ,[12, 7, -25, 6, "%:0"] // front
                                                            ,[12, 7, 182, 6, "100"] // back
                                                            ]);



// ---------- keayboard/keypad ----------
// get the keyboard and connect it to entry field - part of the UI definition
// {"uiKbd3x10B"}.connect(              ui,x,  y, w, h,bc,fc,sc,mx,my,fs,tc,tx,ty,cb) {
var kbd = require("uiKbd3x10B").connect(ui,0,210,24,24, 7, 7,-2,-1,-1,10 ,0, 4, 6,function(c){
    var s = ui.g("i1");
    if (c.length === 1) {
      ui.u("i1",s + "" + c);
    } else if (c === "bs") {
      if (s.length > 0) { ui.u("i1",s.substr(0,s.length - 1)); } 
    }
  });


// ---------- more buttons ----------
//   0  1     2       3  4     5    6    7   8  9       10  11 
// flgs clazz id      x  y     w    h   bc  fc  valObj  cb,  l (label array obj)
//      btn                 ->x2 ->y2                         fv tc    x  y  label text
ui.c(3,"btn","b3"  ,  3,287,  30,  30,  4,   4,     u,  cb);
ui.c(3,"btn","b4"  , 37,287,  30,  30,  2,   2,     u,  cb);
ui.c(3,"btn","b5"  , 71,287,  30,  30,  1,   1,     u,  cb);
ui.c(3,"btn","b6"  ,105,287,  30,  30,  7,   7,     u,  cb);
ui.c(3,"btn","b7"  ,139,287,  30,  30,  6,   6,     u,  cb);
ui.c(3,"btn","b8"  ,173,287,  30,  30,  5,   5,     u,  cb);
ui.c(3,"btn","b9"  ,207,287,  30,  30,  3,   3,     u,  cb);

// var s2=ui.e("s2"); // get and hold on to slider w/ id "s2" for direct handling later on


// ----- run UI
function onInit() {
  SPI1.setup({sck:B3, miso:B4, mosi:B5, baud: 1000000});
                   // spi, dc, cs, rst, callback
  dsp = dMod.connect(SPI1, B7, B6, A8, function() {
      dsp.clear(); A5.set(); // display clear and back light on
      // connect ui to dsp and display it
      ui.connect(dsp)
        .w(0,0,dW-1,dH-1) // wipe (clear/fill) screen with screen background color
        .d() // display all elements
        .di=true; // set display immediately (display element immediately after creation)
      // setup touchscreen and connect it to ui
      SPI2.setup({sck:B13, miso:B14, mosi:B15, baud: 2000000});
      touch = tMod.connect(
          // spi, cs, irq, callback, calc
          SPI2, B10, B1, function(x,y) { ui.evt(x,y);
          }, function(yr, xr, d) {
            return [ Math.round(xr / -121.44          + 259.70685111989) 
                   , Math.round(yr /   88.90357142857 + -19.78130398103)
                   ];
      });
  });
}

onInit();
